from matplotlib_backend_qtquick.qt_compat import QtCore
import matplotlib.pyplot as plt
from anesthetic import make_2d_axes
from .samples_model import Legend
import numpy as np



class TrianglePlotter(QtCore.QObject):
    notify = QtCore.Signal(str)
    paramsChanged = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.beta = 0
        self.logL = -10000
        self.triCanvas = None
        self.higCanvas = None
        self._sample = None
        self.legends = {}
        self.samples = dict()
        self.paramsModel = None
        self._betaCache = {}
        self._LCache = {}

    @property
    def params(self):
        return list(self.paramsModel.displayNames)

    @property
    def tex(self):
        return self.paramsModel.tex

    def _update_triangle(self, post=None):
        if post is None:
            post = lambda x: x
        figsize = self.triCanvas.figure.get_figwidth(), self.triCanvas.figure.get_figheight()
        fig = updateTrianglePlot(plt.figure(figsize=figsize), self.params, self.tex,
                                 self.samples, self.legends, post)
        self.triCanvas.figure = fig
        fig.set_canvas(self.triCanvas)
        self.triCanvas.draw_idle()
        # self._create_higson()

    def _update_higson(self):
        fig = self.higCanvas.figure
        fig.clear()
        ax = fig.gca()
        ax.set_xlabel(r'$\log X$')
        ax.set_ylabel(r'$LX$')
        for x in self.samples:
            with np.errstate(divide='ignore'):
                logX = np.log(self.samples[x].nlive / (self.samples[x].nlive+1)).cumsum()
            LX = self.samples[x].logL/self.beta + logX
            LX = np.exp(LX-LX.max())
            ax.plot(logX, LX, color=self.legends[x].color)
        self.higCanvas.draw_idle()

    @QtCore.Slot(float)
    def changeLogL(self, logL, *args):
        self._LCache[self.logL] = self.triCanvas.figure
        self.logL = logL
        if self.logL in self._LCache:
            self.triCanvas.figure = self._LCache[self.logL]
            self._LCache[self.logL].set_canvas(self.triCanvas)
            self.triCanvas.draw_idle()
        else:
            self._update_triangle(lambda x: x.live_points(self.logL))
            self._LCache[self.logL] = self.triCanvas.figure

    @QtCore.Slot(float)
    def changeTemperature(self, beta, *args):
        self.beta = beta
        self._update_triangle()
        self._update_higson()

    @QtCore.Slot()
    @QtCore.Slot(object)
    @QtCore.Slot(object, object)
    def reDraw(self, samples=None, legends=None, *args):
        if samples is not None:
            self.samples = samples
        if legends is None:
            if self.legends is None:
                for k in self.samples:
                    self.legends[k] = Legend(title=k)
        else:
            self.legends = legends
        self.notify.emit('Full repaint...')
        self._update_triangle()
        self._update_higson()
        self.notify.emit('Fully repainted.')


def updateTrianglePlot(figure, params, tex, samples, legends, postprocess):
    figure.clear()
    figure, axes = make_2d_axes(params, tex=tex, fig=figure, upper=False)
    for x in samples:
        postprocess(samples[x])\
                  .plot_2d(axes,
                           alpha=legends[x].alpha,
                           color=legends[x].color,
                           label=legends[x].title,
                           types={
                               'lower': 'scatter',
                               'diagonal': 'hist',
                           })
    handles, labels = axes[params[0]][params[0]]\
        .get_legend_handles_labels()
    figure.legend(handles, labels)
    return figure



#!/usr/bin/env python
"""
An example of using the backend
"""
import sys
from pathlib import Path
from matplotlib_backend_qtquick.backend_qtquickagg import FigureCanvasQtQuickAgg
from matplotlib_backend_qtquick.qt_compat import QtGui, QtQml, QtCore
from .samples_model import SamplesModel, ParameterModel
from .plt_managers import TrianglePlotter


def main():
    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
    sys.argv += ['--style', 'material']
    app = QtGui.QGuiApplication(sys.argv)
    engine = QtQml.QQmlApplicationEngine()

    context = engine.rootContext()
    displayBridge = TrianglePlotter()
    paramModel = ParameterModel()
    samplesModel = SamplesModel()
    displayBridge.paramsModel = paramModel

    samplesModel.fullRepaint.connect(displayBridge.reDraw)
    samplesModel.newParams.connect(paramModel.addParams)

    context.setContextProperty("displayBridge", displayBridge)
    context.setContextProperty('paramsModel', paramModel)
    context.setContextProperty("samplesModel", samplesModel)

    QtQml.qmlRegisterType(FigureCanvasQtQuickAgg,
                          "Backend", 1, 0, "FigureCanvas")
    qmlFile = Path(Path.cwd(), Path(__file__).parent, "view.qml")
    engine.load(QtCore.QUrl.fromLocalFile(str(qmlFile)))
    win = engine.rootObjects()[0]
    displayBridge.notify.connect(win.displayPythonMessage)
    samplesModel.notify.connect(win.displayPythonMessage)

    displayBridge.triCanvas = win.findChild(QtCore.QObject, "trianglePlot")
    displayBridge.higCanvas = win.findChild(QtCore.QObject, "higsonPlot")
    temperatureSlider = win.findChild(QtCore.QObject, "temperature_slider")
    temperatureSlider.valueChangeFinished\
                     .connect(displayBridge.changeTemperature)
    temperatureSlider.valueChangeStarted\
                     .connect(displayBridge.changeTemperature)
    loglSlider = win.findChild(QtCore.QObject, "logl_slider")
    loglSlider.valueChangeStarted.connect(displayBridge.changeLogL)
    app.exec_()


if __name__ == "__main__":
    main()



"""This module defines the abstracListModel subclass that is used to
keep the QML model in sync with the Python back end.

"""
from matplotlib_backend_qtquick.qt_compat import QtCore
from anesthetic.samples import NestedSamples, MCMCSamples
from os.path import splitext, basename
import matplotlib.pyplot as plt


class Legend:
    colorCycles = plt.rcParams["axes.prop_cycle"].by_key()["color"]
    currentColor = 0

    def __init__(self, title='', color=None, alpha=None):
        self.title = title
        if color is not None:
            self.color = color  # Let's be international.
        else:
            self.color = Legend.colorCycles[Legend.currentColor
                                            % len(Legend.colorCycles)]
        if alpha is None:
            self.alpha = 1
        Legend.currentColor += 1

        @property
        def color(self):
            return self._color

        @color.setter
        def color_(self, color):
            if isinstance(color, str):
                if len(color > 7):
                    self._color = '#{}'.format(color[3:])
                    self.alpha = int(color[1:3], 16)/int("ff", 16)


class ParameterModel(QtCore.QAbstractListModel):
    nameRole = QtCore.Qt.UserRole + 1000 + 0
    texRole = QtCore.Qt.UserRole + 1000 + 1
    selectedRole = QtCore.Qt.UserRole + 1000 + 2

    def __init__(self, parent=None, columns=[], tex={}):
        super(ParameterModel, self).__init__(parent)
        self.names = columns
        self.tex = tex
        self.displayNames = {}

    def roleNames(self):
        roles = {
            ParameterModel.nameRole: b'name',
            ParameterModel.texRole: b'tex',
            ParameterModel.selectedRole: b'selected',
        }
        return roles

    def rowCount(self, parent=QtCore.QModelIndex()):
        if parent.isValid():
            return 0
        return len(self.names)

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if 0 <= index.row() < self.rowCount() and index.isValid():
            if role == ParameterModel.nameRole:
                return self.names[index.row()]
            if role == ParameterModel.texRole:
                return self.parent.tex[self.names[index.row()]]
            if role == ParameterModel.selectedRole:
                return self.names[index.row()] in self.displayNames

    @QtCore.Slot(str)
    @QtCore.Slot(str, str)
    @QtCore.Slot(str, str, bool)
    def appendRow(self, name, tex=None, show=False):
        self.names.add(name)

    @QtCore.Slot(object, object)
    def addParams(self, params, tex):
        self.beginResetModel()
        if self.names == []:
            self.names = params
        else:
            self.names.intersection([params])
            self.tex = {}
        for name in self.names:
            self.tex[name] = tex[name]
        if self.displayNames == {}:
            self.displayNames = set(self.names[:3])
        self.endResetModel()

    def flags(self, index):
        return QtCore.Qt.ItemIsEditable | \
               QtCore.Qt.ItemIsEnabled | \
               QtCore.Qt.ItemIsSelectable

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        if role == ParameterModel.texRole:
            self.tex[self.names[index.row()]] = value
            self.dataChanged.emit(index, index)
            return True
        elif role == ParameterModel.selectedRole:
            name = self.names[index.row()]
            if name in self.displayNames and value \
               or name not in self.displayNames and not value:
                return False
            else:
                if value:
                    self.displayNames.add(name)
                else:
                    self.displayNames.remove(name)
                self.dataChanged.emit(index, index)
                return True
        else:
            return False


class SamplesModel(QtCore.QAbstractListModel):
    fullRepaint = QtCore.Signal(object, object)
    notify = QtCore.Signal(str)
    newParams = QtCore.Signal(object, object)

    nameRole = QtCore.Qt.UserRole + 1000 + 0
    urlRole = QtCore.Qt.UserRole + 1000 + 1
    legendNameRole = QtCore.Qt.UserRole + 1000 + 2
    displayRole = QtCore.Qt.UserRole + 1000 + 3
    legendColorRole = QtCore.Qt.UserRole + 1000 + 4
    samplesTypeRole = QtCore.Qt.UserRole + 1000 + 5
    legendAlphaRole = QtCore.Qt.UserRole + 1000 + 6
    logZRole = QtCore.Qt.UserRole + 1000 + 7
    dklRole = QtCore.Qt.UserRole + 1000 + 8
    bmdRole = QtCore.Qt.UserRole + 1000 + 9

    def roleNames(self):
        roles = {
            SamplesModel.nameRole: b'name',
            SamplesModel.urlRole: b'url',
            SamplesModel.legendNameRole: b'legend_name',
            SamplesModel.legendColorRole: b'legend_color',
            SamplesModel.samplesTypeRole: b'samples_type',
            SamplesModel.displayRole: b'display',
            SamplesModel.legendAlphaRole: b'legend_alpha',
            SamplesModel.logZRole: b'logZ',
            SamplesModel.dklRole: b'Dkl',
            SamplesModel.bmdRole: b'bmd',
        }
        return roles

    def rowCount(self, parent=QtCore.QModelIndex()):
        if parent.isValid():
            return 0
        if len(self.names) != len(self.samples):
            raise ValueError(
                "len(samples) {} and len(samples_names) {}, mismatch."
                .format(len(self.names), len(self.samples)))
        return len(self.samples)

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if 0 <= index.row() < self.rowCount() and index.isValid():
            item = self.samples[self.names[index.row()]]

            if role == SamplesModel.nameRole:
                return self.names[index.row()]

            if role == SamplesModel.urlRole:
                return item.root

            if role == SamplesModel.legendNameRole:
                return self.legends[self.names[index.row()]].title

            if role == SamplesModel.legendColorRole:
                return self.legends[self.names[index.row()]].color

            if role == SamplesModel.displayRole:
                return self.names[index.row()] in self.displayed_names

            if role == SamplesModel.samplesTypeRole:
                if isinstance(item, NestedSamples):
                    return 'NestedSamples'
                if isinstance(item, MCMCSamples):
                    return 'MCMCSamples'
            if role == SamplesModel.legendAlphaRole:
                return self.legends[self.names[index.row()]].alpha

            if role == SamplesModel.bmdRole:
                return float(self.samples[self.names[index.row()]].d())
            if role == SamplesModel.dklRole:
                return float(self.samples[self.names[index.row()]].D())
            if role == SamplesModel.logZRole:
                return float(self.samples[self.names[index.row()]].logZ())

    @QtCore.Slot(str)
    def appendRow(self, file_root, *args):
        rt, _ = cleanupFileRoot(file_root)
        self.notify.emit('Loading...')
        if basename(rt) not in self.names:

            self.beginInsertRows(QtCore.QModelIndex(),
                                 self.rowCount(), self.rowCount())
            samples = NestedSamples(root=rt)
            if samples is None:
                raise ValueError('Samples were None')
            self.names.append(basename(rt))
            self.legends[basename(rt)] = Legend(basename(rt))
            self.samples[basename(rt)] = samples
            self.newParams.emit(samples.columns, samples.tex)
            self.displayed_names.add(basename(rt))
            self.endInsertRows()

            self.reqRepaint()
            self.notify.emit('Loaded.')
        else:
            self.notify.emit(f'Samples: {basename(rt)} - already loaded.')

    def flags(self, index):
        return QtCore.Qt.ItemIsEditable | \
               QtCore.Qt.ItemIsEnabled | \
               QtCore.Qt.ItemIsSelectable

    def reqRepaint(self):
        samples = {k: self.samples[k] for k in self.displayed_names}
        legends = {k: self.legends[k] for k in self.displayed_names}
        self.fullRepaint.emit(samples, legends)

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        if role == SamplesModel.legendNameRole:
            self.legends[self.names[index.row()]].title = value
            self.dataChanged.emit(index, index)
            self.reqRepaint()
            return True
        if role == SamplesModel.legendColorRole:
            self.legends[self.names[index.row()]].color = value.name()
            self.legends[self.names[index.row()]].alpha = value.alphaF()
            self.dataChanged.emit(index, index)
            self.reqRepaint()
            return True
        if role == SamplesModel.legendAlphaRole:
            self.legends[self.names[index.row()]].alpha = value
            self.dataChanged.emit(index, index)
            self.reqRepaint()
            return True
        if role == SamplesModel.displayRole:
            name = self.names[index.row()]
            if name in self.displayed_names and value \
               or (name not in self.displayed_names) and not value:
                return False
            else:
                if value:
                    self.displayed_names.add(name)
                else:
                    self.displayed_names.remove(name)
                self.reqRepaint()
                return True
        else:
            return False

    def __init__(self, parent=None, names=[], samples={}):
        super(SamplesModel, self).__init__(parent)
        self.names = names
        self.samples = samples
        self.legends = {}
        self.displayed_names = set()
        self.parameters = {}


def cleanupFileRoot(file_root):
    ret = file_root.replace('file://', '', 1)
    exts = ['.stats', '.resume', '.paramnames', '.inputparams', '.ranges']
    # Make sure to put the substrings later, so that the longer part can be picked up.
    ends = ['_equal_weights', '_dead-birth', '_dead','_phys_live-birth',  '_phys_live']
    root, ext = splitext(ret)
    if ext in exts:
        return root, ext
    elif ext == '.txt':
        for end in ends:
            try:
                rt, end = root.rsplit(end, 1)
                return rt, end+ext
            except ValueError:
                pass



#!/usr/bin/env python3
from setuptools import setup, find_packages


setup(
    name='paranestamol',
    version='0.0.2',
    description='A QtQuick-based nested sampling visualisation tool.',
    author='Aleksandr Petrosyan',
    author_email='a-p-petrosyan@yandex.ru',
    url='https://github.com/appetrosyan/paranestamol',
    packages=find_packages(),
    install_requires=['matplotlib_backend_qtquick', 'anesthetic', 'pyside2',
                      'numpy', 'matplotlib'],
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License'
    ]
)
